// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
package de.uka.ilkd.key.gui;

import java.util.Properties;
import java.util.LinkedList;
import java.util.Iterator;

/** 
 * This class encapsulates the information about the maximum number of lines
 * a tooltip is allowed to have with instantiated SchemaVariables.
 * If this number is exceeded no SchemaVariabled get instantiated in the 
 * displayed tooltip.
 */
public class ViewSettings implements Settings {


    private static final String MAX_TOOLTIP_LINES_KEY = "[View]MaxTooltipLines";

    /** default max number of displayed tooltip lines is 40 */
    private int maxTooltipLines = 40;


    private LinkedList listenerList = new LinkedList();


    /**
     * @return the current maxTooltipLines
     */
    public int maxTooltipLines() {
    	return maxTooltipLines;
    }
    
    /**
     * Sets maxTooltipLines 
     * @param b The new value for maxTooltipLines 
     */
    public void setMaxTooltipLines(int b) {
		maxTooltipLines = b;
		fireSettingsChanged();
    }

    /** gets a Properties object and has to perform the necessary
     * steps in order to change this object in a way that it
     * represents the stored settings
     * @param The aforementioned Properties object
     */
    public void readSettings(Properties props) {
	String val = props.getProperty(MAX_TOOLTIP_LINES_KEY);
	if (val != null) {
	    setMaxTooltipLines(Integer.valueOf(val).intValue());
	} else {
	    setMaxTooltipLines(maxTooltipLines);
	}
    }


    /** implements the method required by the Settings interface. The
     * settings are written to the given Properties object. Only entries of the form 
     * <key> = <value> (,<value>)* are allowed.
     * @param props the Properties object where to write the settings as (key, value) pair
     */
    public void writeSettings(Properties props) {
	props.setProperty(MAX_TOOLTIP_LINES_KEY, "" + maxTooltipLines);
    }

    /** sends the message that the state of this setting has been
     * changed to its registered listeners (not thread-safe)
     */
    protected void fireSettingsChanged() {
	Iterator it = listenerList.iterator();
	while (it.hasNext()) {
	    ((SettingsListener)it.next()).settingsChanged(new GUIEvent(this));
	}
    }

    /** 
     * adds a listener to the settings object 
     * @param l the listener
     */
    public void addSettingsListener(SettingsListener l) {
	listenerList.add(l);
    }

}
