// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

/* -*-antlr-*- */
header {

package de.uka.ilkd.key.parser;

import java.io.*;
import antlr.*;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.rule.*;
import de.uka.ilkd.key.proof.RuleSource;
import de.uka.ilkd.key.proof.init.Includes;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.logic.sort.*;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.java.abstraction.KeYJavaType;
import de.uka.ilkd.key.java.visitor.*;
import java.util.List;
import java.util.LinkedList;
import java.util.HashMap;
import java.util.HashSet;

}

/** 
 * A parser for signature declarations,
 * i.e. sorts, schema variabls, predicates, functions and heuristics.
 */  
class DeclParser extends Parser;
options {
    importVocab=KeYLexer;
    k = 1;
    defaultErrorHandler=false;
}

{
    static final Sort[] AN_ARRAY_OF_SORTS = new Sort[0];

    private Prioritization.StaticPrioritization heur_prioritization
        = new Prioritization.StaticPrioritization ();
    
    private NamespaceSet nss;
    private Choice defaultChoice = null;
    protected HashMap category2Default = new HashMap();
    protected boolean onlyWith=false;
    protected SetOfChoice activatedChoices = SetAsListOfChoice.EMPTY_SET;
    protected HashSet usedChoiceCategories = new HashSet();

    private Services services;

    // these variables are set if a file is read in step by
    // step. This used when reading in LDTs because of cyclic
    // dependencies.
    protected boolean skip_schemavariables;
    protected boolean skip_functions;
    protected boolean skip_predicates;
    protected boolean skip_sorts;
    protected boolean skip_heuristics;
    private boolean schemaMode = false;
    protected boolean parse_includes = false;
    protected Includes includes = new Includes();

    public DeclParser(TokenStream lexer,
                      String filename,
                      Services services,
                      NamespaceSet nss,
                      Choice defaultChoice) {
        this(lexer);
        setFilename(filename);
        this.nss = nss;
        this.services = services;
        setSkips(false, false, false, false, false);
        switchToNormalMode();
        this.defaultChoice = defaultChoice;
    }

    public DeclParser(TokenStream lexer,
                      String filename,
                      Services services,
                      NamespaceSet nss) {
        this(lexer, filename, services, nss, null);
    }
    
    public JavaInfo getJavaInfo() {
        return getServices().getJavaInfo();
    }

    public Services getServices() {
        return services;
    }

    public Namespace sorts() {
        return nss.sorts();
    }

    public Namespace functions() {
        return nss.functions();
    }

    public Namespace heuristics() {
        return nss.heuristics();
    }

    public Namespace variables() {
        return nss.variables();
    }

    public Namespace programVariables() {
        return nss.programVariables();
    }

    public Namespace choices(){
        return nss.choices();
    }

    public NamespaceSet namespaces() {
        return nss;
    }
    
    public HashMap getCategory2Default(){
        return category2Default;
    }

    private void switchToSchemaMode() {
        schemaMode = true;      
    }

    private boolean inSchemaMode() {
        return schemaMode;
    }

    private void switchToNormalMode() {
        schemaMode = false;  
    }

    public Prioritization getHeuristicPrioritization () {
        return heur_prioritization;
    }

    protected int getLine() {
        int line = -1;
        try {
            line = LT(0).getLine();
        } 
        catch (TokenStreamException e) {
            System.err.println("No further token in stream");
        }
        return line;
    }   

    protected int getColumn() {
        int col = -1;
        try {
            col = LT(0).getColumn();
        } 
        catch (TokenStreamException e) {
            System.err.println("No further token in stream");
        }
        return col;
    }   

    protected void setSkips(boolean svars, boolean funcs, 
                            boolean preds, boolean sorts,
                            boolean heurs) {
        skip_schemavariables = svars;
        skip_functions       = funcs;
        skip_predicates      = preds;
        skip_sorts       = sorts;
        skip_heuristics      = heurs;
    }

    public Named lookup(Name n) {
	final Namespace[] lookups = {
		functions(), variables(), 
		programVariables()
	};
	for (int i = 0; i<lookups.length; i++) {
	    if (lookups[i].lookup(n) != null) {
		return lookups[i].lookup(n);
            }
	}
	return null;	
    }

    public void parseVariables()    
        throws antlr.RecognitionException, antlr.TokenStreamException
        {
            setSkips(false, true, true, true, true); 
            top();
            setSkips(false, false, false, false, false);                
        }

    protected void addInclude(String filename, boolean relativePath, boolean ldt){
        InputStream input=null;
        RuleSource source=null;
        if (relativePath) {
            int end = getFilename().lastIndexOf(File.separator);
            int start = 0;
            filename = filename.replace('/', File.separatorChar);
            filename = filename.replace('\\', File.separatorChar);
            if(getFilename().startsWith("file:")){
                start = 5;
            }
            File path=new File(getFilename().substring(start,end+1)+filename);
            try{ 
                source = RuleSource.initRuleFile(path.toURL()); 
            }catch(java.net.MalformedURLException e){
                System.err.println("Exception due to malformed URL of file "+
                                   filename+"\n " +e);
            }
            filename = filename.substring(
                                          filename.lastIndexOf(File.separator)+1,
                                          filename.lastIndexOf(".key"));
        } else {
            source = RuleSource.initRuleFile(filename+".key"); 
        }
        if (ldt) {
            includes.putLDT(filename, source);
        } else {
            includes.put(filename, source);
        }
    }  

    public void parseFunctions()
        throws antlr.RecognitionException, antlr.TokenStreamException

        {
            setSkips(true, false, true, true, true); 
            top();
            setSkips(false, false, false, false, false);
        }

    public void parsePredicates()
        throws antlr.RecognitionException, antlr.TokenStreamException
        {
            setSkips(true, true, false, true, true); 
            top();
            setSkips(false, false, false, false, false);        
        }

    public void parseSorts() 
        throws antlr.RecognitionException, antlr.TokenStreamException
        {
            setSkips(true, true, true, false, true); 
            top();
            setSkips(false, false, false, false, false);
        }

    public void parseHeuristics()
        throws antlr.RecognitionException, antlr.TokenStreamException
        {
            setSkips(true, true, true, true, false); 
            top();
            setSkips(false, false, false, false, false);        
        }

    public void parseFuncAndPred()
        throws antlr.RecognitionException, antlr.TokenStreamException
        {
            setSkips(true, false, false, true, true); 
            top();
            setSkips(false, false, false, false, false);
        }

}

top : decls;

decls : 

        (INCLUDE 
            ( LDTS COLON LBRACE include[true] (COMMA include[true])* RBRACE
            |
                include[false]
            )
            SEMI)*
        {
            if(parse_includes) return;
            activatedChoices = SetAsListOfChoice.EMPTY_SET;  
        }
        ("with" LPAREN activated_choice (COMMA activated_choice)* RPAREN)?
        {if(onlyWith) return;}
        (
            option_decls
        |    
            sort_decls
        |
            prog_var_decls
        |
            schema_var_decls
        |
            pred_decls
        |
            func_decls
        |
            heur_decls

        ) *
    ;

include [boolean ldt]:
        (absfile:IDENT{ 
                if(parse_includes){
                    addInclude(absfile.getText(),false,ldt);
                }
            }
        | relfile:STRING_LITERAL{ 
                if(parse_includes){
                    addInclude(relfile.getText(),true,ldt);
                }
            })
    ;

activated_choice{
    String name;
    Choice c;
}:
        cat:IDENT COLON choice:IDENT
        {if(usedChoiceCategories.contains(cat.getText())){
            throw new IllegalArgumentException("You have already chosen a different option for "+cat.getText());
        }
        usedChoiceCategories.add(cat.getText());
        name = cat.getText()+":"+choice.getText();
        c = (Choice) choices().lookup(new Name(name));
        if(c==null){
            throw new NotDeclException("Option", choice,
                                       getFilename());
        }else{
            activatedChoices=activatedChoices.add(c);
            namespaces().setFunctions(functions().extended(c.funcNS().allElements()));
        }
        }
    ;

option_decls{
}:
        "options" LBRACE choice (SEMI choice)* RBRACE 
    ;

choice{
    String cat=null;
}:
        category:IDENT {cat=category.getText();} (COLON LBRACE choice_option[cat] (COMMA choice_option[cat])* RBRACE)? 
        {
            if(!category2Default.containsKey(cat)){
                choices().add(new Choice("On",cat));
                choices().add(new Choice("Off",cat)); 
                category2Default.put(cat, cat+":On");               
            }
        }
    ;

choice_option[String cat]{
    String name;
}:
        choice:IDENT { name=cat+":"+choice.getText();
        Choice c = (Choice) choices().lookup(new Name(name));
        if(c==null){
            c = new Choice(choice.getText(),cat);
            choices().add(c);
        }
        if(!category2Default.containsKey(cat)){
            category2Default.put(cat, name);
        }
        }
    ;

sort_decls {
    boolean isObjectSort=false;
    boolean isGenericSort=false;
    Sort[] sortExt=new Sort [0];
    Sort[] sortOneOf=new Sort [0];
    ListOfName sortIds;
}:
        "sorts" 
        LBRACE ( 
            { 
                isObjectSort=false; 
                isGenericSort=false; 
                sortExt = new Sort[0];
                sortOneOf = new Sort[0];
            }
            ( "object" {isObjectSort=true;} sortIds = sortIdentifiers
            | "generic" {isGenericSort=true;} sortIds = sortIdentifiers
                ( "oneof" sortOneOf = oneof_sorts )? 
                ("extends" sortExt = extends_sorts )?
            | sortIds = sortIdentifiers
            )
            SEMI
            {   
                if (!skip_sorts) {
                    Name sort_name;
                    IteratorOfName it = sortIds.iterator ();
                    while ( it.hasNext () ) {
                        sort_name = it.next ();

                        if (sorts().lookup(sort_name) == null) {
                            Sort s;
                            if (isObjectSort) {
                                if ("Object".equals(sort_name)) {
                                    if (sorts().lookup(new Name("Object")) == null) {
                                        s = new ClassInstanceSortImpl(sort_name);
                                    }
                                    s=(Sort)sorts().lookup(new Name("Object"));
                                } else {
                                    s = new ClassInstanceSortImpl(sort_name, 
                                                                  (Sort)sorts().lookup
                                                                  (new Name("Object")));
                                }   
                            } else if (isGenericSort) {
                                int        i;
                                SetOfSort  ext   = GenericSort.EMPTY_SORT_SET;
                                SetOfSort  oneOf = GenericSort.EMPTY_SORT_SET;

                                for ( i = 0; i != sortExt.length; ++i )
                                    ext = ext.add ( sortExt[i] );

                                for ( i = 0; i != sortOneOf.length; ++i )
                                    oneOf = oneOf.add ( sortOneOf[i] );

                                try {
                                    s = new GenericSort ( sort_name,
                                                          ext,
                                                          oneOf );
                                } catch ( GenericSupersortException e ) {
                                    throw new GenericSortException ( "sort",
                                                                     "Illegal sort given", e.getIllegalSort (),
                                                                     getFilename (), getLine (), getColumn () );
                                }
                            } else {
                                s = new PrimitiveSort(sort_name);
                            }
                            sorts().add(s);
                            if (s instanceof NonCollectionSort) {
                                NonCollectionSort ns
                                    = (NonCollectionSort)s;
                                sorts().add(ns.getSetSort());
                                sorts().add(ns.getSequenceSort());
                                sorts().add(ns.getBagSort());
                            }
                            if ( s instanceof SortDefiningSymbols )
                                ((SortDefiningSymbols)s).createSymbols ( defaultChoice.funcNS(),
                                                                         sorts() );
                        }

                    }
                }
            }
            
        )* 
        RBRACE
    ;

sortIdentifiers returns [ListOfName ids = SLListOfName.EMPTY_LIST]
    :
        id:IDENT { ids = ids.append ( new Name(id.getText()) ); }
        (
            COMMA
            id2:IDENT { ids = ids.append ( new Name(id2.getText()) ); }
        ) *
    ;

extends_sorts returns [Sort[] extendsSorts = null] 
{
    List args = new LinkedList();
    Sort s;
}
    :
        s = any_sortId_check[true] { args.add(s); }
        (
            COMMA s = any_sortId_check[true] {args.add(s);}
        ) *
        {
            extendsSorts = (Sort[])args.toArray(AN_ARRAY_OF_SORTS);
        }
    ;

oneof_sorts returns [Sort[] oneOfSorts = null] 
{
    List args = new LinkedList();
    Sort s;
}
    :
        LBRACE
        s = sortId_check[true] { args.add(s); }
        (
            COMMA s = sortId_check[true] {args.add(s);}
        ) *
        RBRACE
        {
            oneOfSorts = (Sort[])args.toArray(AN_ARRAY_OF_SORTS);
        }
    ;

prog_var_decls 
{
    String var_name;
    String type = null;
}
    :
        { switchToNormalMode(); }
        "program" "variables"
        LBRACE 
        (
            ((v0:IDENT {type= "" + v0.getText();} (DOT v1:IDENT {type= type + "." + v1.getText();})*)
              |
             (pid:PIDENT {type = pid.getText();} )
            )
            (LBRACKET RBRACKET {
               type = type + "[]";
            })?
            //sortId_check[true]
            var_name = id
            {
                KeYJavaType kjt = getJavaInfo().getKeYJavaType(type);
		ProgramElementName pvName = new ProgramElementName(var_name);
                if (kjt == null) {
                    throw new antlr.SemanticException
                        ("Program Variable " + var_name + 
                         " declared with unknown type " + 
                         type);

                } else if (lookup(pvName)!=null) {
		  // commented out as pv do not have unique name (at the moment)
		    //throw new AmbigiousDeclException
		//	(var_name, getFilename(), getLine(), getColumn()); 
		}
            namespaces().programVariables().add(new ProgramVariable
                (new ProgramElementName(var_name), kjt));

            }
            SEMI
        ) *  
        RBRACE
    ;

schema_var_decls
{
    Sort s = null;
    boolean makeVariableSV  = false;
    boolean makeDependingSV = false; // skolem functions
    boolean rigid           = false;
    boolean listModifier    = false;
}
    :      
        "schema" "variables"
        LBRACE  { switchToSchemaMode(); }
        (
            ( "modal" "operator" one_schema_modal_decl (COMMA one_schema_modal_decl)* SEMI )
        |
            {makeVariableSV = false; makeDependingSV = false; rigid = false; listModifier=false;}
            ( ( "program" ("svlist" {listModifier=true;})?
                    ( "statement"
                        {
                            s=ProgramSVSort.STATEMENT;
                        }
                    | "expression"
                        {
                            s=ProgramSVSort.EXPRESSION;
                        }
                    | "javaboolean" "expression"
                        {
                            s=ProgramSVSort.JAVABOOLEANEXPRESSION;
                        }
                    | "executioncontext"
                        {
                            s=ProgramSVSort.EXECUTIONCONTEXT;
                        }
                    | "multiplevardecl"
                        {
                            s=ProgramSVSort.MULTIPLEVARDECL;
                        }
                    | "simple"
                        ( "expression"
                            {
                                s=ProgramSVSort.SIMPLEEXPRESSION;
                            }
                        | "javabyte" "expression"
                            {
                                s=ProgramSVSort.SIMPLEJAVABYTEEXPRESSION;
                            }
                        | "javachar" "expression"
                            {
                                s=ProgramSVSort.SIMPLEJAVACHAREXPRESSION;
                            }
                        | "javashort" "expression"
                            {
                                s=ProgramSVSort.SIMPLEJAVASHORTEXPRESSION;
                            }
                        |  "javaint" "expression"
                            {
                                s=ProgramSVSort.SIMPLEJAVAINTEXPRESSION;
                            }
                        |  "javalong" "expression"
                            {
                                s=ProgramSVSort.SIMPLEJAVALONGEXPRESSION;
                            }
                        |  "javabyteshortint" "expression"
                            {
                                s=ProgramSVSort.SIMPLEJAVABYTESHORTINTEXPRESSION;
                            }
                        | "anyjavatype" "expression"
                            {
                                s=ProgramSVSort.SIMPLEANYJAVATYPEEXPRESSION;
                            }
                        )
                    | "nonsimple" 
                        ("expression"   
                            {
                                s=ProgramSVSort.NONSIMPLEEXPRESSION;
                            }
                        |
                            "methodreference" {
                                s=ProgramSVSort.NONSIMPLEMETHODREFERENCE;
                            }) 
                    | "loopinit"
                        {
                            s=ProgramSVSort.LOOPINIT;
                        }
                    | "guard"
                        {
                            s=ProgramSVSort.GUARD;
                        }
                    | "forupdates"
                        {
                            s=ProgramSVSort.FORUPDATES;
                        }
                    | "label"
                        {
                            s=ProgramSVSort.LABEL;
                        }
                    | "type"
                        {
                            s=ProgramSVSort.TYPE;
                        }
                    | "notprimitive" "type" 
                        {
                            s=ProgramSVSort.TYPENOTPRIMITIVE;
                        }
                    | "explicit" "variable" 
                        {
                            s=ProgramSVSort.EXPLICITVARIABLE;
                        }
                    | "static" "variable"   
                        {
                            s=ProgramSVSort.STATICVARIABLE;
                        }
                    | "variable"    
                        {
                            s=ProgramSVSort.VARIABLE;
                        }
                    | "constant" "variable" 
                        {
                            s=ProgramSVSort.CONSTANTVARIABLE;
                        } 
                    | "variableinitializer"
                        {
                            s=ProgramSVSort.VARIABLEINIT;
                        }
                    | "implicit" 
                        (  "variable"   
                            {
                                s=ProgramSVSort.IMPLICITVARIABLE;
                            }
                        | "reference" "field"   
                            {
                                s=ProgramSVSort.IMPLICITREFERENCE;
                            }               
                        | id:IDENT 
                            {
                                HashMap name2sort = new HashMap(10);
                                name2sort.put("next", ProgramSVSort.IMPLICITNEXT);
                                name2sort.put("nextToCreate", ProgramSVSort.IMPLICITNEXTTOCREATE);
                                name2sort.put("classInitialized", ProgramSVSort.IMPLICITCLINIT);
                                name2sort.put("classInitializationInProgress", 
                                              ProgramSVSort.IMPLICITINITINPROGRESS);
                                name2sort.put("classErroneous", ProgramSVSort.IMPLICITERRONEOUS);
                                name2sort.put("classPrepared", ProgramSVSort.IMPLICITPREPARED);                                                             
                                s = (Sort) name2sort.get(id.getText());
                                if (s == null) {
                                    throw new antlr.SemanticException
                                        ("SchemaVariable of sort "+id.getText()+ " not found.");
                                }
                            }
                        )   
                    | "begin" "transaction"     
                        {
                            s=ProgramSVSort.JCBEGINTRANSACTION;
                        }               
                    | "commit" "transaction" 
                        {
                            s=ProgramSVSort.JCCOMMITTRANSACTION;
                        }               
                    | "abort" "transaction" 
                        {
                            s=ProgramSVSort.JCABORTTRANSACTION;
                        }               
                    | "transient" 
                        ( "booleanarray" 
                            {
                                s=ProgramSVSort.JCMAKETRANSIENTBOOLEANARRAY;
                            }               
                        |  "bytearray" 
                            {
                                s=ProgramSVSort.JCMAKETRANSIENTBYTEARRAY;
                            }               
                        |  "shortarray" 
                            {
                                s=ProgramSVSort.JCMAKETRANSIENTSHORTARRAY;
                            }               
                        |  "objectarray" 
                            {
                                s=ProgramSVSort.JCMAKETRANSIENTOBJECTARRAY;
                            }               
                        |  "istransient" 
                            {
                                s=ProgramSVSort.JCISTRANSIENT;
                            }       
                        )       
                    | "array" (
                            {"initializer".equals(LT(1).getText()) || 
                             "length".equals(LT(1).getText()) }? 
                             arrayId:IDENT { 
                                if ( "initializer".equals(arrayId.getText()) ) {
                                    s=ProgramSVSort.ARRAYINITIALIZER; 
                                } else { 
                                    s=ProgramSVSort.ARRAYLENGTH; 
                                }                       
                            } 
                        )                            
                    | "new" ( "array" { s=ProgramSVSort.NEWARRAY; }
                        |  { s=ProgramSVSort.NEW; } )
                    | "switch"
                        {
                            s=ProgramSVSort.SWITCH;
                        }
                    | "catch" {
                            s=ProgramSVSort.CATCH;
                        }
                    | "methodname" {
                            s=ProgramSVSort.METHODNAME;
                        }
                    |
                        "special" "constructorreference" {
                            s=ProgramSVSort.SPECIALCONSTRUCTORREFERENCE;                
                        }   
                    | "methodbody" {
                            s=ProgramSVSort.METHODBODY;
                        }
                    | "methodreference" {
                            s=ProgramSVSort.METHODREFERENCE;
                        }                                      
                    | id1:IDENT {
                            if ("lefthandside".equals(id1.getText())) {
                                s=ProgramSVSort.LEFTHANDSIDE;
                            } 
                        }
                    )             
                )
            |   
                "formula" 
                ( "rigid" { rigid = true; } ) ?
                { 
                    s = Sort.FORMULA;
                }
            |
                s = any_sortId_check[true] 
                ( "variables" { makeVariableSV = true; }
                | "depending" { makeDependingSV = true; }
                | "rigid" { rigid = true; } ) ?             
            )
            one_schema_var_decl[s,makeVariableSV,makeDependingSV,rigid,listModifier]
            ( COMMA one_schema_var_decl[s,makeVariableSV,makeDependingSV,rigid,listModifier] ) *
            SEMI
        ) *
        { switchToNormalMode(); }
        RBRACE
    ;

one_schema_var_decl[Sort s, 
                    boolean makeVariableSV, 
                    boolean makeDependingSV,
                    boolean rigid, 
                    boolean listModifier]
    :
        id:IDENT
        {
            if (!skip_schemavariables) {
                SchemaVariable v;
                if ( s == Sort.FORMULA ) {
                    v = SchemaVariableFactory.createFormulaSV
                        (new Name(id.getText()), listModifier, rigid);
                } else if ( s instanceof ProgramSVSort ) {
                    v = SchemaVariableFactory.createProgramSV
                        (new ProgramElementName(id.getText()),
				(ProgramSVSort) s, listModifier);
                } else {
                    if ( makeVariableSV ) {
                        v=SchemaVariableFactory.createVariableSV
                            (new Name(id.getText()), s, listModifier);
                    } else if ( makeDependingSV ) {
                        v = SchemaVariableFactory.createDependingSV
                            (new Name(id.getText()), s, listModifier);
                    } else { v = SchemaVariableFactory.createTermSV
                                 (new Name(id.getText()), s, listModifier, rigid);
                    }
                }
                if (inSchemaMode()) {		    		
                    if ( lookup(v.name()) != null ) {
			throw new AmbigiousDeclException
			   (v.name().toString(), getFilename(), getLine(), getColumn());
		    } else {
			variables().add(v);
		    }
                }
            }
        }
    ; 


one_schema_modal_decl
{
    Modality[] mods;
} 
    :
        id:IDENT EQUALS mods=modalities_decl {
            SchemaVariable msv = (SchemaVariable)variables().lookup(new Name(id.getText()));
            if(msv != null)
            throw new antlr.SemanticException("Schema variable "+id.getText()+ " already defined");
            int arity = mods[0].arity();
            for(int i=1;i<mods.length;i++)
            if(arity != mods[i].arity())
            throw new antlr.SemanticException("Arity mismatch for schema variable "+id.getText());
            msv = SchemaVariableFactory.createModalOperatorSV(new Name(id.getText()), arity, mods);
            if (inSchemaMode()) {
                variables().add(msv);
            }
        }
    ;

modalities_decl returns [Modality[] modalities = null] 
{
    List args = new LinkedList();
}
    : (LBRACE 
            s1:IDENT {
                if(s1.getText().charAt(0) == '#') {
                    ModalOperatorSV msv = (ModalOperatorSV)variables().lookup(new Name(s1.getText()));
                    if(msv == null)
                        throw new antlr.SemanticException("Schema variable "+s1.getText()+ " not defined");
                    for(int i=0; i<msv.modalities().length; i++)
                        args.add(msv.modalities()[i]);
                }else{
                    Modality op = Op.getModality(s1.getText());
                    if(op == null)
                        throw new antlr.SemanticException("Unrecognised modal operator "+s1.getText());
                    args.add(op);
                }
            } (COMMA s2:IDENT {
                    if(s2.getText().charAt(0) == '#') {
                        ModalOperatorSV msv = (ModalOperatorSV)variables().lookup(new Name(s2.getText()));
                        if(msv == null)
                            throw new antlr.SemanticException("Schema variable "+s2.getText()+ " not defined");
                        for(int i=0; i<msv.modalities().length; i++)
                            args.add(msv.modalities()[i]);
                    } else {
                        Modality op = Op.getModality(s2.getText());
                        if(op == null)
                            throw new antlr.SemanticException("Unrecognised modal operator "+s2.getText());
                        args.add(op);
                    }
                })* RBRACE
        ) {
            modalities = (Modality[])args.toArray(new Modality[args.size()]);
        }
    ;

pred_decls 
{
    Sort[] argSorts;
    String pred_name;
    Namespace oldFNS=null;
    String ctext = null;
    Choice c=null;
}
    :
        "predicates" (LPAREN cat:IDENT COLON choice:IDENT RPAREN)?
        { if(cat!=null){
            ctext = cat.getText()+":"+choice.getText();
            c = (Choice) choices().lookup(new Name(ctext));
            if(c==null){
                throw new NotDeclException("Option", choice,
                                           getFilename());
            }else{
                namespaces().setFunctions(c.funcNS());
            }
        }else{
            namespaces().setFunctions(defaultChoice.funcNS());
        }
        cat=null;
        }
        LBRACE
        (
            pred_name = id
            argSorts=arg_sorts[!skip_predicates]
            {
                if (!skip_predicates) {
                    Name predicate = new Name(pred_name);		    		    
                    if (lookup(predicate) != null) {
                        throw new AmbigiousDeclException
                        (pred_name, getFilename(), getLine(), getColumn());
                    } else {
                        Function p = new Function(predicate,
                                                  Sort.FORMULA, 
                                                  argSorts);
                        functions().add(p);         
                    }
                } 
            }
            SEMI
        ) *
        RBRACE
    ;

func_decls 
{
    Sort[] argSorts;
    Sort retSort;
    String func_name;
    Namespace oldFNS=null;
    String ctext = null;
    Choice c=null;
}
    :
        "functions" (LPAREN cat:IDENT COLON choice:IDENT RPAREN)?
        { 
            if(cat!=null){
                ctext = cat.getText()+":"+choice.getText();
                c = (Choice) choices().lookup(new Name(ctext));
                if(c==null){
                    throw new NotDeclException("Option", choice,
                                               getFilename());
                }
            }else{
                c=defaultChoice;
            }
            namespaces().setFunctions(c.funcNS());
            cat=null;
        }
        LBRACE
        (
            retSort = sortId_check[!skip_functions]
            func_name = id 
            argSorts = arg_sorts[!skip_functions]
            {
                if (!skip_functions) {
                    Name fct_name = new Name(func_name);
                    if (lookup(fct_name) != null) {
                        throw new AmbigiousDeclException
                        (func_name, getFilename(), getLine(), getColumn());
                    } else {
                        functions().add(new Function(fct_name,
     	                                             retSort, 
        	                                     argSorts));
                    }
                } 
            }
            SEMI
        ) *
        RBRACE
    ;

arg_sorts[boolean checkSort] returns [Sort[] argSorts = null] 
{
    List args = new LinkedList();
    Sort s;
}
    :
        (
            LPAREN
            s = sortId_check[checkSort] { args.add(s); }
            (
                COMMA s = sortId_check[checkSort] {args.add(s);}
            ) *
            RPAREN
        ) ?
        {
            argSorts = (Sort[])args.toArray(AN_ARRAY_OF_SORTS);
        }
    ;


integer_num returns [ int num = 0 ]
options {defaultErrorHandler=false;}
    :
        {
            try {
                Token t = LT(1);
                num = Integer.decode ( t.getText () ).intValue ();
                consume ();
            } catch ( TokenStreamException e ) {
                throw new antlr.SemanticException ( "Expected a number",
                                                    getFilename (), getLine (), getColumn () );
            } catch ( NumberFormatException e ) {
                throw new antlr.SemanticException ( "Expected a number",
                                                    getFilename (), getLine (), getColumn () );
            } catch ( NullPointerException e ) {
                throw new antlr.SemanticException ( "Expected a number",
                                                    getFilename (), getLine (), getColumn () );
            }
        }
    ;

heur_decls
{
    int     t_priority = 0;
    Integer priority   = null;
}
    :
        "heuristics"
        LBRACE
        (
            { priority = null; }
            id:IDENT
            (
                LPAREN t_priority = integer_num RPAREN 
                { priority = new Integer ( t_priority ); }
            ) ?
            SEMI
            { 
                if (!skip_heuristics) {
                    Heuristic h = new Heuristic(new Name(id.getText()));
                    if(heuristics().lookup(new Name(id.getText()))==null){
                        heuristics().add(h);
                    }
                    if ( priority != null )
                        heur_prioritization.setPriority ( h, priority.intValue () );
                }
            }
        ) *
        RBRACE
    ;

sortId  returns [Sort s = null]
    :
        s = sortId_check[true]
    ;           

// Non-generic sorts, array sorts allowed
sortId_check [boolean checkSort] returns [Sort s = null]                
{
    Sort t;
}
    :
        t = sortId_check_help[checkSort]
        s = array_decls[t]
    ;

// Generic and non-generic sorts, array sorts allowed
any_sortId_check [boolean checkSort] returns [Sort s = null]                
{
    Sort t;
}
    :
        t = any_sortId_check_help[checkSort]
        s = array_decls[t]
    ;

// also allow generic sorts
any_sortId_check_help [boolean checkSort] returns [Sort s = null]
{
    String name;
}
    :
        name = id
        {   s= (Sort) sorts().lookup(new Name(name));
            if (checkSort && s == null) {
                throw new NotDeclException("sort", name, 
                                           getFilename(), 
                                           getLine(), getColumn());
            }
        }
    ;

sortId_check_help [boolean checkSort] returns [Sort s = null]
    :
        s = any_sortId_check_help[checkSort]
        {
            // don't allow generic sorts or collection sorts of
            // generic sorts at this point
            Sort t = s;
            while ( t instanceof CollectionSort )
            t = ((CollectionSort)t).elementSort ();
            if ( t instanceof GenericSort )
            throw new GenericSortException ( "sort",
                                             "Non-generic sort expected", t,
                                             getFilename (), getLine (), getColumn () );
        }
    ;

array_decls [Sort p] returns [Sort s = null]                
{
    int  n=0;
}

    :
        (LBRACKET RBRACKET {n++;})*

        { 
            if(n != 0){
                s = ArraySortImpl.getArraySortForDim(
                    p, n, getJavaInfo().getObjectSort(),
                    getJavaInfo().getCloneableSort());
            }else{
                s = p;
            }
        }
            
    ;

id_declaration returns [ IdDeclaration idd = null ]
{
    Sort s = null;
}
    :
        id:IDENT
        ( COLON s = sortId_check[true] ) ?
        {
            idd = new IdDeclaration ( id.getText (), s );
        }
    ;

id returns [String name = null]
    :
        (prefix:IDENT { name = prefix.getText(); }
        | pprefix : PIDENT { name = pprefix.getText(); }
        )
        ((COLON COLON)=>COLON COLON nameID:IDENT 
            { 
                name += "::" + nameID.getText();
            }
        )?
    ;

